#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License
#

"""
PROTON-2121 python-qpid-proton 0.28 BlockingConnection leaks connections (does not close file descriptors)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib
import gc
import os
import socket
import subprocess
import threading
import unittest
import warnings

import cproton

import proton.handlers
import proton.utils
import proton.reactor


def get_fd_set():
    # type: () -> set[str]
    return set(os.listdir('/proc/self/fd/'))


@contextlib.contextmanager
def no_fd_leaks(test):
    # type: (unittest.TestCase) -> None
    with warnings.catch_warnings(record=True) as ws:
        before = get_fd_set()
        yield
        delta = get_fd_set().difference(before)
        if len(delta) != 0:
            subprocess.check_call("ls -lF /proc/{0}/fd/".format(os.getpid()), shell=True)
            test.fail("Found {0} new fd(s) after the test".format(delta))

        if len(ws) > 0:
            test.fail([w.message for w in ws])


class Broker(proton.handlers.MessagingHandler):
    def __init__(self, acceptor_url):
        # type: (str) -> None
        super(Broker, self).__init__()
        self.acceptor_url = acceptor_url

        self.acceptor = None
        self._acceptor_opened_event = threading.Event()

    def get_acceptor_sockname(self):
        # type: () -> (str, int)
        self._acceptor_opened_event.wait()
        if hasattr(self.acceptor, '_selectable'):  # proton 0.30.0+
            sockname = self.acceptor._selectable._delegate.getsockname()
        else:  # works in proton 0.27.0
            selectable = cproton.pn_cast_pn_selectable(self.acceptor._impl)
            fd = cproton.pn_selectable_get_fd(selectable)
            s = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
            sockname = s.getsockname()
        return sockname[:2]

    def on_start(self, event):
        self.acceptor = event.container.listen(self.acceptor_url)
        self._acceptor_opened_event.set()

    def on_link_opening(self, event):
        if event.link.is_sender:
            assert not event.link.remote_source.dynamic, "This cannot happen"
            event.link.source.address = event.link.remote_source.address
        elif event.link.remote_target.address:
            event.link.target.address = event.link.remote_target.address


@contextlib.contextmanager
def test_broker():
    broker = Broker('localhost:0')
    container = proton.reactor.Container(broker)
    threading.Thread(target=container.run).start()

    try:
        yield broker
    finally:
        container.stop()


PROC_SELF_FD_EXISTS = os.path.exists("/proc/self/fd"), "Skipped: Directory /proc/self/fd does not exist"


class BlockingConnectionFDLeakTests(unittest.TestCase):
    @unittest.skipUnless(*PROC_SELF_FD_EXISTS)
    @unittest.expectedFailure
    def test_just_start_stop_test_broker(self):
        with no_fd_leaks(self):
            with test_broker() as broker:
                broker.get_acceptor_sockname()  # wait for acceptor to open

            gc.collect()

    @unittest.skipUnless(*PROC_SELF_FD_EXISTS)
    @unittest.expectedFailure
    def test_connection_close_all(self):
        with no_fd_leaks(self):
            with test_broker() as broker:
                c = proton.utils.BlockingConnection("{0}:{1}".format(*broker.get_acceptor_sockname()))
                c.close()

            gc.collect()

    @unittest.skipUnless(*PROC_SELF_FD_EXISTS)
    def test_connection_close_all__do_not_check_test_broker(self):
        with test_broker() as broker:
            acceptor_sockname = broker.get_acceptor_sockname()
            with no_fd_leaks(self):
                c = proton.utils.BlockingConnection("{0}:{1}".format(*acceptor_sockname))
                c.close()

                gc.collect()

    @unittest.skipUnless(*PROC_SELF_FD_EXISTS)
    @unittest.expectedFailure
    def test_connection_sender_close_all(self):
        with no_fd_leaks(self):
            with test_broker() as broker:
                c = proton.utils.BlockingConnection("{0}:{1}".format(*broker.get_acceptor_sockname()))
                s = c.create_sender("anAddress")
                s.close()
                c.close()

            gc.collect()

    @unittest.skipUnless(*PROC_SELF_FD_EXISTS)
    @unittest.expectedFailure
    def test_connection_receiver_close_all(self):
        with no_fd_leaks(self):
            with test_broker() as broker:
                c = proton.utils.BlockingConnection("{0}:{1}".format(*broker.get_acceptor_sockname()))
                s = c.create_receiver("anAddress")
                s.close()
                c.close()

            gc.collect()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

import sys

from proton import *

from . import common


class Test(common.Test):
    pass


class ClientTransportTest(Test):

    def setUp(self):
        self.transport = Transport()
        self.peer = Transport()
        self.conn = Connection()
        self.peer.bind(self.conn)

    def tearDown(self):
        self.transport = None
        self.peer = None
        self.conn = None

    def drain(self):
        while True:
            p = self.transport.pending()
            if p < 0:
                return
            elif p > 0:
                data = self.transport.peek(p)
                self.peer.push(data)
                self.transport.pop(len(data))
            else:
                assert False

    def assert_error(self, name):
        assert self.conn.remote_container is None, self.conn.remote_container
        self.drain()
        # verify that we received an open frame
        assert self.conn.remote_container is not None, self.conn.remote_container
        # verify that we received a close frame
        assert self.conn.state == Endpoint.LOCAL_UNINIT | Endpoint.REMOTE_CLOSED, self.conn.state
        # verify that a framing error was reported
        assert self.conn.remote_condition.name == name, self.conn.remote_condition

    def testEOS(self):
        self.transport.push(b"")  # should be a noop
        self.transport.close_tail()  # should result in framing error
        self.assert_error(u'amqp:connection:framing-error')

    def testPartial(self):
        self.transport.push(b"AMQ")  # partial header
        self.transport.close_tail()  # should result in framing error
        self.assert_error(u'amqp:connection:framing-error')

    def testGarbage(self, garbage=b"GARBAGE_"):
        self.transport.push(garbage)
        self.assert_error(u'amqp:connection:framing-error')
        assert self.transport.pending() < 0
        self.transport.close_tail()
        assert self.transport.pending() < 0

    def testSmallGarbage(self):
        self.testGarbage(b"XXX")

    def testBigGarbage(self):
        self.testGarbage(b"GARBAGE_XXX")

    def testHeader(self):
        self.transport.push(b"AMQP\x00\x01\x00\x00")
        self.transport.close_tail()
        self.assert_error(u'amqp:connection:framing-error')

    def testHeaderBadDOFF1(self):
        """Verify doff > size error"""
        self.testGarbage(b"AMQP\x00\x01\x00\x00\x00\x00\x00\x08\x08\x00\x00\x00")

    def testHeaderBadDOFF2(self):
        """Verify doff < 2 error"""
        self.testGarbage(b"AMQP\x00\x01\x00\x00\x00\x00\x00\x08\x01\x00\x00\x00")

    def testHeaderBadSize(self):
        """Verify size > max_frame_size error"""
        self.transport.max_frame_size = 512
        self.testGarbage(b"AMQP\x00\x01\x00\x00\x00\x00\x02\x01\x02\x00\x00\x00")

    def testProtocolNotSupported(self):
        self.transport.push(b"AMQP\x01\x01\x0a\x00")
        p = self.transport.pending()
        assert p >= 8, p
        bytes = self.transport.peek(p)
        assert bytes[:8] == b"AMQP\x00\x01\x00\x00"
        self.transport.pop(p)
        self.drain()
        assert self.transport.closed

    def testPeek(self):
        out = self.transport.peek(1024)
        assert out is not None

    def testBindAfterOpen(self):
        conn = Connection()
        ssn = conn.session()
        conn.open()
        ssn.open()
        conn.container = "test-container"
        conn.hostname = "test-hostname"
        trn = Transport()
        trn.bind(conn)
        out = trn.peek(1024)
        assert b"test-container" in out, repr(out)
        assert b"test-hostname" in out, repr(out)
        self.transport.push(out)

        c = Connection()
        assert c.remote_container is None
        assert c.remote_hostname is None
        assert c.session_head(0) is None
        self.transport.bind(c)
        assert c.remote_container == "test-container"
        assert c.remote_hostname == "test-hostname"
        assert c.session_head(0) is not None

    def testCloseHead(self):
        n = self.transport.pending()
        assert n > 0, n
        try:
            self.transport.close_head()
        except TransportException:
            e = sys.exc_info()[1]
            assert "aborted" in str(e), str(e)
        n = self.transport.pending()
        assert n < 0, n

    def testCloseTail(self):
        n = self.transport.capacity()
        assert n > 0, n
        try:
            self.transport.close_tail()
        except TransportException:
            e = sys.exc_info()[1]
            assert "aborted" in str(e), str(e)
        n = self.transport.capacity()
        assert n < 0, n

    def testUnpairedPop(self):
        conn = Connection()
        self.transport.bind(conn)

        conn.hostname = "hostname"
        conn.open()

        dat1 = self.transport.peek(1024)

        ssn = conn.session()
        ssn.open()

        dat2 = self.transport.peek(1024)

        assert dat2[:len(dat1)] == dat1

        snd = ssn.sender("sender")
        snd.open()

        self.transport.pop(len(dat1))
        self.transport.pop(len(dat2) - len(dat1))
        dat3 = self.transport.peek(1024)
        self.transport.pop(len(dat3))
        assert self.transport.peek(1024) == b""

        self.peer.push(dat1)
        self.peer.push(dat2[len(dat1):])
        self.peer.push(dat3)


class ServerTransportTest(Test):

    def setUp(self):
        self.transport = Transport(Transport.SERVER)
        self.peer = Transport()
        self.conn = Connection()
        self.peer.bind(self.conn)

    def tearDOwn(self):
        self.transport = None
        self.peer = None
        self.conn = None

    def drain(self):
        while True:
            p = self.transport.pending()
            if p < 0:
                return
            elif p > 0:
                bytes = self.transport.peek(p)
                self.peer.push(bytes)
                self.transport.pop(len(bytes))
            else:
                assert False

    def assert_error(self, name):
        assert self.conn.remote_container is None, self.conn.remote_container
        self.drain()
        # verify that we received an open frame
        assert self.conn.remote_container is not None, self.conn.remote_container
        # verify that we received a close frame
        assert self.conn.state == Endpoint.LOCAL_UNINIT | Endpoint.REMOTE_CLOSED, self.conn.state
        # verify that a framing error was reported
        assert self.conn.remote_condition.name == name, self.conn.remote_condition

    # TODO: This may no longer be testing anything
    def testEOS(self):
        self.transport.push(b"")  # should be a noop
        self.transport.close_tail()
        p = self.transport.pending()
        self.drain()
        assert self.transport.closed

    def testPartial(self):
        self.transport.push(b"AMQ")  # partial header
        self.transport.close_tail()
        p = self.transport.pending()
        assert p >= 8, p
        bytes = self.transport.peek(p)
        assert bytes[:8] == b"AMQP\x00\x01\x00\x00"
        self.transport.pop(p)
        self.drain()
        assert self.transport.closed

    def testGarbage(self, garbage=b"GARBAGE_"):
        self.transport.push(garbage)
        p = self.transport.pending()
        assert p >= 8, p
        bytes = self.transport.peek(p)
        assert bytes[:8] == b"AMQP\x00\x01\x00\x00"
        self.transport.pop(p)
        self.drain()
        assert self.transport.closed

    def testSmallGarbage(self):
        self.testGarbage(b"XXX")

    def testBigGarbage(self):
        self.testGarbage(b"GARBAGE_XXX")

    def testHeader(self):
        self.transport.push(b"AMQP\x00\x01\x00\x00")
        self.transport.close_tail()
        self.assert_error(u'amqp:connection:framing-error')

    def testProtocolNotSupported(self):
        self.transport.push(b"AMQP\x01\x01\x0a\x00")
        p = self.transport.pending()
        assert p >= 8, p
        bytes = self.transport.peek(p)
        assert bytes[:8] == b"AMQP\x00\x01\x00\x00"
        self.transport.pop(p)
        self.drain()
        assert self.transport.closed

    def testPeek(self):
        out = self.transport.peek(1024)
        assert out is not None

    def testBindAfterOpen(self):
        conn = Connection()
        ssn = conn.session()
        conn.open()
        ssn.open()
        conn.container = "test-container"
        conn.hostname = "test-hostname"
        trn = Transport()
        trn.bind(conn)
        out = trn.peek(1024)
        assert b"test-container" in out, repr(out)
        assert b"test-hostname" in out, repr(out)
        self.transport.push(out)

        c = Connection()
        assert c.remote_container is None
        assert c.remote_hostname is None
        assert c.session_head(0) is None
        self.transport.bind(c)
        assert c.remote_container == "test-container"
        assert c.remote_hostname == "test-hostname"
        assert c.session_head(0) is not None

    def testCloseHead(self):
        n = self.transport.pending()
        assert n >= 0, n
        try:
            self.transport.close_head()
        except TransportException:
            e = sys.exc_info()[1]
            assert "aborted" in str(e), str(e)
        n = self.transport.pending()
        assert n < 0, n

    def testCloseTail(self):
        n = self.transport.capacity()
        assert n > 0, n
        try:
            self.transport.close_tail()
        except TransportException:
            e = sys.exc_info()[1]
            assert "aborted" in str(e), str(e)
        n = self.transport.capacity()
        assert n < 0, n

    def testUnpairedPop(self):
        conn = Connection()
        self.transport.bind(conn)

        conn.hostname = "hostname"
        conn.open()

        dat1 = self.transport.peek(1024)

        ssn = conn.session()
        ssn.open()

        dat2 = self.transport.peek(1024)

        assert dat2[:len(dat1)] == dat1

        snd = ssn.sender("sender")
        snd.open()

        self.transport.pop(len(dat1))
        self.transport.pop(len(dat2) - len(dat1))
        dat3 = self.transport.peek(1024)
        self.transport.pop(len(dat3))
        assert self.transport.peek(1024) == b""

        self.peer.push(dat1)
        self.peer.push(dat2[len(dat1):])
        self.peer.push(dat3)

    def testEOSAfterSASL(self):
        self.transport.sasl().allowed_mechs('ANONYMOUS')

        self.peer.sasl().allowed_mechs('ANONYMOUS')

        # this should send over the sasl header plus a sasl-init set up
        # for anonymous
        p = self.peer.pending()
        self.transport.push(self.peer.peek(p))
        self.peer.pop(p)

        # now we send EOS
        self.transport.close_tail()

        # the server may send an error back
        p = self.transport.pending()
        while p > 0:
            self.peer.push(self.transport.peek(p))
            self.transport.pop(p)
            p = self.transport.pending()

        # server closed
        assert self.transport.pending() < 0


class LogTest(Test):

    def testTracer(self):
        t = Transport()
        assert t.tracer is None
        messages = []

        def tracer(transport, message):
            messages.append((transport, message))
        t.tracer = tracer
        assert t.tracer is tracer
        t.log("one")
        t.log("two")
        t.log("three")
        assert messages == [(t, "TRACE: one"), (t, "TRACE: two"), (t, "TRACE: three")], messages



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

from cproton import PN_DEFAULT_PRIORITY, PN_STRING, PN_UUID, PN_OVERFLOW, pn_error_text, pn_message, \
    pn_message_annotations, pn_message_body, pn_message_clear, pn_message_decode, \
    pn_message_encode, pn_message_error, pn_message_free, pn_message_get_address, pn_message_get_content_encoding, \
    pn_message_get_content_type, pn_message_get_correlation_id, pn_message_get_creation_time, pn_message_get_delivery_count, \
    pn_message_get_expiry_time, pn_message_get_group_id, pn_message_get_group_sequence, pn_message_get_id, pn_message_get_priority, \
    pn_message_get_reply_to, pn_message_get_reply_to_group_id, pn_message_get_subject, pn_message_get_ttl, \
    pn_message_get_user_id, pn_message_instructions, pn_message_is_durable, \
    pn_message_is_first_acquirer, pn_message_is_inferred, pn_message_properties, pn_message_set_address, \
    pn_message_set_content_encoding, pn_message_set_content_type, pn_message_set_correlation_id, pn_message_set_creation_time, \
    pn_message_set_delivery_count, pn_message_set_durable, pn_message_set_expiry_time, pn_message_set_first_acquirer, \
    pn_message_set_group_id, pn_message_set_group_sequence, pn_message_set_id, pn_message_set_inferred, pn_message_set_priority, \
    pn_message_set_reply_to, pn_message_set_reply_to_group_id, pn_message_set_subject, \
    pn_message_set_ttl, pn_message_set_user_id

from ._common import isinteger, millis2secs, secs2millis, unicode2utf8, utf82unicode
from ._data import char, Data, symbol, ulong, AnnotationDict
from ._endpoints import Link
from ._exceptions import EXCEPTIONS, MessageException
from uuid import UUID
from typing import Dict, Optional, Union, TYPE_CHECKING, overload

if TYPE_CHECKING:
    from proton._delivery import Delivery
    from proton._endpoints import Sender, Receiver
    from proton._data import Described, PythonAMQPData


class Message(object):
    """The :py:class:`Message` class is a mutable holder of message content.

    :ivar instructions: delivery instructions for the message ("Delivery Annotations" in the AMQP 1.0 spec)
    :vartype instructions: ``dict``
    :ivar ~.annotations: infrastructure defined message annotations ("Message Annotations" in the AMQP 1.0 spec)
    :vartype ~.annotations: ``dict``
    :ivar ~.properties: application defined message properties
    :vartype ~.properties: ``dict``
    :ivar body: message body

    :param kwargs: Message property name/value pairs to initialize the Message
    """

    DEFAULT_PRIORITY = PN_DEFAULT_PRIORITY
    """ Default AMQP message priority"""

    def __init__(
            self,
            body: Union[bytes, str, dict, list, int, float, 'UUID', 'Described', None] = None,
            **kwargs
    ) -> None:
        self._msg = pn_message()
        self.instructions = None
        self.annotations = None
        self.properties = None
        self.body = body
        for k, v in kwargs.items():
            getattr(self, k)  # Raise exception if it's not a valid attribute.
            setattr(self, k, v)

    def __del__(self) -> None:
        if hasattr(self, "_msg"):
            pn_message_free(self._msg)
            del self._msg

    def _check(self, err: int) -> int:
        if err < 0:
            exc = EXCEPTIONS.get(err, MessageException)
            raise exc("[%s]: %s" % (err, pn_error_text(pn_message_error(self._msg))))
        else:
            return err

    def _check_property_keys(self) -> None:
        """
        AMQP allows only string keys for properties. This function checks that this requirement is met
        and raises a MessageException if not. However, in certain cases, conversions to string are
        automatically performed:

        1. When a key is a user-defined (non-AMQP) subclass of str.
           AMQP types symbol and char, although derived from str, are not converted,
           and result in an exception.
        """
        # We cannot make changes to the dict while iterating, so we
        # must save and make the changes afterwards
        changed_keys = []
        for k in self.properties.keys():
            if isinstance(k, str):
                # strings and their subclasses
                if type(k) is symbol or type(k) is char:
                    # Exclude symbol and char
                    raise MessageException('Application property key is not string type: key=%s %s' % (str(k), type(k)))
                if type(k) is not str:
                    # Only for string subclasses, convert to string
                    changed_keys.append((k, str(k)))
            else:
                # Anything else: raise exception
                raise MessageException('Application property key is not string type: key=%s %s' % (str(k), type(k)))
        # Make the key changes
        for old_key, new_key in changed_keys:
            self.properties[new_key] = self.properties.pop(old_key)

    def _pre_encode(self) -> None:
        inst = Data(pn_message_instructions(self._msg))
        ann = Data(pn_message_annotations(self._msg))
        props = Data(pn_message_properties(self._msg))
        body = Data(pn_message_body(self._msg))

        inst.clear()
        if self.instructions is not None:
            inst.put_object(self.instructions)
        ann.clear()
        if self.annotations is not None:
            ann.put_object(self.annotations)
        props.clear()
        if self.properties is not None:
            self._check_property_keys()
            props.put_object(self.properties)
        body.clear()
        if self.body is not None:
            body.put_object(self.body)

    def _post_decode(self) -> None:
        inst = Data(pn_message_instructions(self._msg))
        ann = Data(pn_message_annotations(self._msg))
        props = Data(pn_message_properties(self._msg))
        body = Data(pn_message_body(self._msg))

        if inst.next():
            self.instructions = inst.get_object()
        else:
            self.instructions = None
        if ann.next():
            self.annotations = ann.get_object()
        else:
            self.annotations = None
        if props.next():
            self.properties = props.get_object()
        else:
            self.properties = None
        if body.next():
            self.body = body.get_object()
        else:
            self.body = None

    def clear(self) -> None:
        """
        Clears the contents of the :class:`Message`. All fields will be reset to
        their default values.
        """
        pn_message_clear(self._msg)
        self.instructions = None
        self.annotations = None
        self.properties = None
        self.body = None

    @property
    def inferred(self) -> bool:
        """The inferred flag for a message indicates how the message content
        is encoded into AMQP sections. If inferred is true then binary and
        list values in the body of the message will be encoded as AMQP DATA
        and AMQP SEQUENCE sections, respectively. If inferred is false,
        then all values in the body of the message will be encoded as AMQP
        VALUE sections regardless of their type.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return pn_message_is_inferred(self._msg)

    @inferred.setter
    def inferred(self, value: bool) -> None:
        self._check(pn_message_set_inferred(self._msg, bool(value)))

    @property
    def durable(self) -> bool:
        """The durable property indicates that the message should be held durably
        by any intermediaries taking responsibility for the message.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return pn_message_is_durable(self._msg)

    @durable.setter
    def durable(self, value: bool) -> None:
        self._check(pn_message_set_durable(self._msg, bool(value)))

    @property
    def priority(self) -> int:
        """The relative priority of the message, with higher numbers indicating
        higher priority. The number of available priorities depends
        on the implementation, but AMQP defines the default priority as
        the value ``4``. See the
        `OASIS AMQP 1.0 standard
        <http://docs.oasis-open.org/amqp/core/v1.0/os/amqp-core-messaging-v1.0-os.html#type-header>`_
        for more details on message priority.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return pn_message_get_priority(self._msg)

    @priority.setter
    def priority(self, value: int) -> None:
        self._check(pn_message_set_priority(self._msg, value))

    @property
    def ttl(self) -> float:
        """The time to live of the message measured in seconds. Expired messages
        may be dropped.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return millis2secs(pn_message_get_ttl(self._msg))

    @ttl.setter
    def ttl(self, value: Union[float, int]) -> None:
        self._check(pn_message_set_ttl(self._msg, secs2millis(value)))

    @property
    def first_acquirer(self) -> bool:
        """``True`` iff the recipient is the first to acquire the message,
        ``False`` otherwise.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return pn_message_is_first_acquirer(self._msg)

    @first_acquirer.setter
    def first_acquirer(self, value: bool) -> None:
        self._check(pn_message_set_first_acquirer(self._msg, bool(value)))

    @property
    def delivery_count(self) -> int:
        """The number of delivery attempts made for this message.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return pn_message_get_delivery_count(self._msg)

    @delivery_count.setter
    def delivery_count(self, value: int) -> None:
        self._check(pn_message_set_delivery_count(self._msg, value))

    @property
    def id(self) -> Optional[Union[str, bytes, 'UUID', ulong]]:
        """The globally unique id of the message, and can be used
        to determine if a received message is a duplicate. The allowed
        types to set the id are:

        :type: The valid AMQP types for an id are one of:

               * ``int`` (unsigned)
               * ``uuid.UUID``
               * ``bytes``
               * ``str``
        """
        value = pn_message_get_id(self._msg)
        if isinstance(value, tuple):
            if value[0] == PN_UUID:
                value = UUID(bytes=value[1])
        return value

    @id.setter
    def id(self, value: Optional[Union[str, bytes, 'UUID', int]]) -> None:
        if isinteger(value):
            value = ulong(value)
        elif isinstance(value, UUID):
            value = (PN_UUID, value.bytes)
        pn_message_set_id(self._msg, value)

    @property
    def user_id(self) -> bytes:
        """The user id of the message creator.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return pn_message_get_user_id(self._msg)

    @user_id.setter
    def user_id(self, value: bytes) -> None:
        self._check(pn_message_set_user_id(self._msg, value))

    @property
    def address(self) -> Optional[str]:
        """The address of the message.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return utf82unicode(pn_message_get_address(self._msg))

    @address.setter
    def address(self, value: str) -> None:
        self._check(pn_message_set_address(self._msg, unicode2utf8(value)))

    @property
    def subject(self) -> Optional[str]:
        """The subject of the message.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return utf82unicode(pn_message_get_subject(self._msg))

    @subject.setter
    def subject(self, value: str) -> None:
        self._check(pn_message_set_subject(self._msg, unicode2utf8(value)))

    @property
    def reply_to(self) -> Optional[str]:
        """The reply-to address for the message.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return utf82unicode(pn_message_get_reply_to(self._msg))

    @reply_to.setter
    def reply_to(self, value: str) -> None:
        self._check(pn_message_set_reply_to(self._msg, unicode2utf8(value)))

    @property
    def correlation_id(self) -> Optional[Union['UUID', ulong, str, bytes]]:
        """The correlation-id for the message.

        :type: The valid AMQP types for a correlation-id are one of:

               * ``int`` (unsigned)
               * ``uuid.UUID``
               * ``bytes``
               * ``str``
        """
        value = pn_message_get_correlation_id(self._msg)
        if isinstance(value, tuple):
            if value[0] == PN_UUID:
                value = UUID(bytes=value[1])
        return value

    @correlation_id.setter
    def correlation_id(self, value: Optional[Union[str, bytes, 'UUID', int]]) -> None:
        if isinteger(value):
            value = ulong(value)
        elif isinstance(value, UUID):
            value = (PN_UUID, value.bytes)
        pn_message_set_correlation_id(self._msg, value)

    @property
    def content_type(self) -> symbol:
        """The RFC-2046 [RFC2046] MIME type for the message body.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return symbol(utf82unicode(pn_message_get_content_type(self._msg)))

    @content_type.setter
    def content_type(self, value: str) -> None:
        self._check(pn_message_set_content_type(self._msg, unicode2utf8(value)))

    @property
    def content_encoding(self) -> symbol:
        """The content-encoding of the message.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return symbol(utf82unicode(pn_message_get_content_encoding(self._msg)))

    @content_encoding.setter
    def content_encoding(self, value: str) -> None:
        self._check(pn_message_set_content_encoding(self._msg, unicode2utf8(value)))

    @property
    def expiry_time(self) -> float:  # TODO doc said int
        """The absolute expiry time of the message in seconds using the Unix time_t [IEEE1003] encoding.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return millis2secs(pn_message_get_expiry_time(self._msg))

    @expiry_time.setter
    def expiry_time(self, value: Union[float, int]) -> None:
        self._check(pn_message_set_expiry_time(self._msg, secs2millis(value)))

    @property
    def creation_time(self) -> float:
        """The creation time of the message in seconds using the Unix time_t [IEEE1003] encoding.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return millis2secs(pn_message_get_creation_time(self._msg))

    @creation_time.setter
    def creation_time(self, value: Union[float, int]) -> None:
        self._check(pn_message_set_creation_time(self._msg, secs2millis(value)))

    @property
    def group_id(self) -> Optional[str]:
        """The group id of the message.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return utf82unicode(pn_message_get_group_id(self._msg))

    @group_id.setter
    def group_id(self, value: str) -> None:
        self._check(pn_message_set_group_id(self._msg, unicode2utf8(value)))

    @property
    def group_sequence(self) -> int:
        """The sequence of the message within its group.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return pn_message_get_group_sequence(self._msg)

    @group_sequence.setter
    def group_sequence(self, value: int) -> None:
        self._check(pn_message_set_group_sequence(self._msg, value))

    @property
    def reply_to_group_id(self) -> Optional[str]:
        """The group-id for any replies.

        :raise: :exc:`MessageException` if there is any Proton error when using the setter.
        """
        return utf82unicode(pn_message_get_reply_to_group_id(self._msg))

    @reply_to_group_id.setter
    def reply_to_group_id(self, value: str) -> None:
        self._check(pn_message_set_reply_to_group_id(self._msg, unicode2utf8(value)))

    @property
    def instructions(self) -> Optional[AnnotationDict]:
        """Delivery annotations as a dictionary of key/values. The AMQP 1.0
        specification restricts this dictionary to have keys that are either
        :class:`symbol` or :class:`ulong` types. It is possible to use
        the special ``dict`` subclass :class:`AnnotationDict` which
        will by default enforce these restrictions on construction. In addition,
        if string types are used, this class will be silently convert them into
        symbols.

        :type: :class:`AnnotationDict`. Any ``dict`` with :class:`ulong` or :class:`symbol` keys.
        """
        return self.instruction_dict

    @instructions.setter
    def instructions(self, instructions: Optional[Dict[Union[str, int], 'PythonAMQPData']]) -> None:
        if isinstance(instructions, dict):
            self.instruction_dict = AnnotationDict(instructions, raise_on_error=False)
        else:
            self.instruction_dict = instructions

    @property
    def annotations(self) -> Optional[AnnotationDict]:
        """Message annotations as a dictionary of key/values. The AMQP 1.0
        specification restricts this dictionary to have keys that are either
        :class:`symbol` or :class:`ulong` types. It is possible to use
        the special ``dict`` subclass :class:`AnnotationDict` which
        will by default enforce these restrictions on construction. In addition,
        if a string types are used, this class will silently convert them into
        symbols.

        :type: :class:`AnnotationDict`. Any ``dict`` with :class:`ulong` or :class:`symbol` keys.
        """
        return self.annotation_dict

    @annotations.setter
    def annotations(self, annotations: Optional[Dict[Union[str, int], 'PythonAMQPData']]) -> None:
        if isinstance(annotations, dict):
            self.annotation_dict = AnnotationDict(annotations, raise_on_error=False)
        else:
            self.annotation_dict = annotations

    def encode(self) -> bytes:
        self._pre_encode()
        sz = 16
        while True:
            err, data = pn_message_encode(self._msg, sz)
            if err == PN_OVERFLOW:
                sz *= 2
                continue
            else:
                self._check(err)
                return data

    def decode(self, data: bytes) -> None:
        self._check(pn_message_decode(self._msg, data))
        self._post_decode()

    def send(self, sender: 'Sender', tag: Optional[str] = None) -> 'Delivery':
        """
        Encodes and sends the message content using the specified sender,
        and, if present, using the specified tag. Upon success, will
        return the :class:`Delivery` object for the sent message.

        :param sender: The sender to send the message
        :param tag: The delivery tag for the sent message
        :return: The delivery associated with the sent message
        """
        dlv = sender.delivery(tag or sender.delivery_tag())
        encoded = self.encode()
        sender.stream(encoded)
        sender.advance()
        if sender.snd_settle_mode == Link.SND_SETTLED:
            dlv.settle()
        return dlv

    @overload
    def recv(self, link: 'Sender') -> None:
        ...

    def recv(self, link: 'Receiver') -> Optional['Delivery']:
        """
        Receives and decodes the message content for the current :class:`Delivery`
        from the link. Upon success it will return the current delivery
        for the link. If there is no current delivery, or if the current
        delivery is incomplete, or if the link is not a receiver, it will
        return ``None``.

        :param link: The link to receive a message from
        :return: the delivery associated with the decoded message (or None)

        """
        if link.is_sender:
            return None
        dlv = link.current
        if not dlv or dlv.partial:
            return None
        dlv.encoded = link.recv(dlv.pending)
        link.advance()
        # the sender has already forgotten about the delivery, so we might
        # as well too
        if link.remote_snd_settle_mode == Link.SND_SETTLED:
            dlv.settle()
        self.decode(dlv.encoded)
        return dlv

    def __repr__(self) -> str:
        props = []
        for attr in ("inferred", "address", "reply_to", "durable", "ttl",
                     "priority", "first_acquirer", "delivery_count", "id",
                     "correlation_id", "user_id", "group_id", "group_sequence",
                     "reply_to_group_id", "instructions", "annotations",
                     "properties", "body"):
            value = getattr(self, attr)
            if value:
                props.append("%s=%r" % (attr, value))
        return "Message(%s)" % ", ".join(props)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

#
# A platform-agnostic tool for running a program in a modified environment.
#

import sys
import os
import subprocess
from optparse import OptionParser


def main(argv=None):

    parser = OptionParser(usage="Usage: %prog [options] [--] VAR=VALUE... command [options] arg1 arg2...")
    parser.add_option("-i", "--ignore-environment",
                      action="store_true", default=False,
                      help="Start with an empty environment (do not inherit current environment)")

    (options, args) = parser.parse_args(args=argv)

    if options.ignore_environment:
        new_env = {}
    else:
        new_env = os.environ.copy()

    # pull out each name value pair
    while len(args):
        z = args[0].split("=", 1)
        if len(z) != 2:
            break  # done with env args
        name, value = z[0], z[1]
        if len(name) == 0:
            raise Exception("Error: incorrect format for env var: '%s'" % str(args[0]))
        del args[0]
        if len(value) == 0:
            # value is not present, so delete it
            if name in new_env:
                del new_env[name]
        else:
            new_env[name] = value

    if len(args) == 0 or len(args[0]) == 0:
        raise Exception("Error: syntax error in command arguments")

    p = subprocess.Popen(args, env=new_env)
    return p.wait()


if __name__ == "__main__":
    sys.exit(main())
